// $Id: Kernel.cc,v 1.4.2.3 1998/10/01 17:09:42 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Thread.h"
#include "Process.h"
#include "Kernel.h"
#include "Gadgets.h"

Kernel TheKernel;


Kernel::Kernel(): theActOffset(0), theCcLevel(0), theCcLevelToReport(100) {
	// we are not playing games here
	theThreads.stretch(4096);
	theActivists.stretch(4096);
}

Kernel::~Kernel() {
	/* disable these to avoid coredumps on ^C */
	//Assert(!theCcLevel);
	//Assert(!theActivists.count());
}

// note: do not run too many threads at once (alarms and such might get late)
void Kernel::slice() {
	Notepad << "start slicing " << theActivists.count() << " threads; offset: " << theActOffset << ende;
	Assert(theActOffset <= theActivists.count());
	for (int i = 0; theActOffset < theActivists.count() && i < 1000; ++i, ++theActOffset) {
		Thread *th = theActivists[theActOffset];
		theActivists[theActOffset] = 0;
		if (th) {
			Notepad << "thread[" << th->kernIdx() << "]: " << th << ' ' << th->name() << " is active, turn: " << theActOffset << ende;
			th->actIdx(-1);
			if (!th->done()) {
				th->checkProgress();
				if (th->runnable())
					th->run();
				if (!th->done())
					continue;
			}
			remove(th);
		}
	}
	Notepad << "stop slicing at " << theActOffset << " threads." << ende;
	if (theActOffset == theActivists.count()) { // reset if sliced all
		theActivists.reset();
		theActOffset = 0;
	}
}

int Kernel::start(Thread *th) {
	int idx = -1;
	if (th) {
		// find idx
		if (!theThreads.full()) {      // greedy but fast
			idx = theThreads.count();
			theThreads.append(th);
		} else
		if (theThreads.find(0, idx)) { // slow linear search
			theThreads[idx] = th;
		} else {
			idx = theThreads.count();  // expensive alloc
			theThreads.append(th);
		}
		th->kernIdx(idx);
		th->init();
		Notepad << "thread[" << idx << "]: " << th << ' ' << th->name() << " started." << ende;

		if (theCcLevel++ >= theCcLevelToReport) {
			SysLog << "number of concurrent threads went over " << theCcLevelToReport << ende;
			theCcLevelToReport += max(100, theCcLevelToReport/10);
		}
	}
	return idx;
}

void Kernel::stopKid(Process *parent, int idx) {
	Assert(parent);
	Assert(idx >= 0);
	Thread *th = theThreads[idx];
	if (th && th->parent() == parent) {
		Notepad << "thread[" << idx << "]: " << th << ' ' << th->name() << " ~stopped as " << parent << " kid." << ende;
		stop(th);
	}
}

void Kernel::stopKids(Process *parent) {
	Assert(parent);
	Notepad << "stopping all kids for parent " << parent << ende;
	for (int i = 0; i < theThreads.count(); ++i) {
		Thread *th = theThreads[i];
		if (th && th->parent() == parent) {
			Notepad << "thread[" << i << "]: " << th << ' ' << th->name() << " ~stopped as " << th->parent() << " kid." << ende;
			stop(th);
		}
	}
}

void Kernel::activate(Thread *th) {
	Assert(th);
	Assert(!th->active());
	Assert(th->kernIdx() >= 0);
	Assert(theThreads[th->kernIdx()] == th);
	th->actIdx(theActivists.count());
	theActivists.append(th);
	Notepad << "thread[" << th->kernIdx() << "]: " << th << ' ' << th->name() << " to be activated, turn: " << th->actIdx() << ende;
}

void Kernel::remove(Thread *th) {
	if (th) {
		if (th->active()) {
			Assert(theActivists[th->actIdx()] == th);
			theActivists[th->actIdx()] = 0;
		}

		const int idx = th->kernIdx();
		Notepad << "thread[" << idx << "]: " << th << ' ' << th->name() << " ~removed." << ende;

		Assert(theThreads[idx] == th);
		theThreads[idx] = 0;
		th->kernIdx(-1);
		th->removed();
		Process *parent = th->parent();
		if (parent) {
			parent->delKid(idx);
			delete th;
			Notepad << "parent[" << parent->kernIdx() << "] " << parent << " ready to die: " << parent->readyToDie() << ende;
			// check if parent is not being removed, but needs to be
			if (parent->kernIdx() >= 0 && parent->readyToDie()) {
				Notepad << "removing ready-to-die parent " << parent << ende;
				stop(parent);
			}
		} else
			Notepad << "thread[" << idx << "]: " << th << ' ' << th->name() << " has no parent" << ende;

		// XXX: hidden hack above: orphan threads do not get deleted

		theCcLevel--;
	}
}
