// $Id: Server.cc,v 1.26.2.5 1998/11/16 18:14:48 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

// #include <netinet/in.h> /* sockaddr_in */

#include "ConnMgr.h"
#include "SvXaction.h"
#include "ConnInfo.h"
#include "AlarmClock.h"
#include "Server.h"
#include "LocalStats.h"
#include "ErrHist.h"
#include "Log.h"

/* ServerCfg */

ServerCfg::ServerCfg(): theConnMgr(0), thePause(0) {
}

void ServerCfg::destroy() {
	delete theConnMgr; theConnMgr = 0;
	delete thePause; thePause = 0;
	theXactCfg.destroy();
}

ServerCfg ServerCfg::clone() const {
	ServerCfg cfg;
	cfg.theConnMgr = theConnMgr->clone();
	cfg.thePause = thePause->clone();
	cfg.theXactCfg = theXactCfg.clone();
	return cfg;
}

/* Server */

Server::Server(): theNextPortIdx(0) {
}

Server::~Server() {
	for (int i = 0; i < theListenSockets.count(); ++i) {
		if (theListenSockets[i] >= 0)
			deaf(i);
	}
	theCfg.destroy();
}

Thread *Server::clone() const {
	Server *s = new Server();
	s->configure(theCfg.clone());
	return s;
}

void Server::configure(const ServerCfg &aCfg) {
	theCfg = aCfg;
	theCfg.theConnMgr->owner(this);
}

// serve several ports
void Server::ports(const IntArray &ports) {
	Journal << "started " << name() << ' ' << this << " will listen on " << ports.count() << " ports" << ende;
	theListenSockets.stretch(ports.count());
	for (int i = 0; i < ports.count(); ++i)
		listen(ports[i]);
}

// serve only one port
void Server::port(int port) {
	Journal << "started " << name() << ' ' << this << " will listen on " << 1 << " port" << ende;
	theListenSockets.stretch(1);
	listen(port);
}

// start listening on a given port
void Server::listen(int port) {
	Socket s;
	if (!s.create()) {
		TheErrHist[errno]++;
		SysLog << "server " << this << " failed to create a listen socket" << ende;
		abort();
	}
	s.blocking(false);
	s.reuseAddr(true);
	s.linger(0);
	theListenSockets.append(s);
	assert(s.listen(port));
	TheSelectLoop.setFD(s, ioRead, this);
	Journal << "server " << name() << ' ' << this << " listens on port " << port << ende;
}

void Server::idleRes(ResMgr *mgr) {
	{ // XXX: KLUDGE to avoid persistent connections
		if (existConn(mgr) && theConns[mgr->resIdx()]->idle())
			closeConn(mgr->resIdx());
		return;
	}

	if (existConn(mgr))
		launchXact(*theConns[mgr->resIdx()]);
}

void Server::voidRes(ResMgr *mgr) {
	if (existConn(mgr)) {
		closeConn(mgr->resIdx());
		StatInc(theXactErrCount);
		theErr.tick();
	}
}

// do we still have a record of this connection?
bool Server::existConn(ResMgr *mgr) const {
	if (!mgr)
		return false;
	const int idx = mgr->resIdx();
	return 
		idx < theConns.count() && 
		theConns[idx] && 
		theConns[idx]->id() == mgr->resId();
}

// close _existing_ connection
void Server::closeConn(int idx) {
	ConnInfo *ci = theConns[idx];
	theConns[idx] = 0;
	assert(ci->conn());
	ci->conn()->close();
	delete ci;
}

// accept a connection
void Server::readReady(int fd) {
	static Time lastCall = TheClock;
	static Time delayRecord;
	static int accRecord = 0;
	static int callCount = 0;

	++callCount;

	const Time t = TheClock;
	const Time d = t - lastCall;

	int count = 0;
	while (accept(fd)) { count++; } // accept all?
	Journal << "server " << this << " accepted: " << count << " conns" << ende;

	if (d > delayRecord && count) {
		cerr << "Accept[" << callCount << "]: record delay of " << d << " accepted: " << count << endl;
		delayRecord = d;
	}
	if (count > accRecord) {
		cerr << "Accept[" << callCount << "]: record accepted count of " << count << endl;
		accRecord = count;
	}
	lastCall = t;
}

// accept a connection
void Server::checkPorts() {
	if (int c = theListenSockets.count())
		readReady(theListenSockets[theNextPortIdx++ % c]);
}

void Server::launchXact(ConnInfo &ci) {
	ci.use();
	Xaction *x = new SvXaction();
	x->configure(theCfg.theXactCfg.clone());
	x->conn(
		ci.getMx(ConnInfo::dirIn),
		ci.getMx(ConnInfo::dirOut));
	startKid(x);
}

// accept pending connection on a given fd if any
bool Server::accept(int fd) {
	bool fatal;
	Connection *conn = new Connection();
	if (conn->accept(fd, fatal)) {
		ConnInfo *ci = new ConnInfo(this, conn);
		int idx;
		if (theConns.find(0, idx))
			theConns[idx] = ci;
		else
			theConns.append(ci);
		ci->idx(idx);
		launchXact(*ci);
		return true;
	}

	// some kind of failure
	delete conn;

	// close the listen socket on fatal errors ?
	if (fatal) {
		int idx;
		assert(theListenSockets.find(fd, idx));
		deaf(idx);
	}
	return false;
}

// un-listen on a socket
void Server::deaf(int idx) {
	Socket s = theListenSockets[idx];
	theListenSockets[idx] = -1;
	TheSelectLoop.clearFD(s, ioRead);
	SysLog << "server " << this << " closed listen socket " << s << ende;
	s.close();
}
