// $Id: globals.cc,v 1.5.2.12 1998/11/17 20:54:56 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"
#include "Log.h"
#include "globals.h"

#include <math.h>
#include <ctype.h>

#include <fstream.h>

World TheWorld;
Opt TheOpt;

int TheMaxFD = FD_SETSIZE; // XXX: this should be a configured value!
const char *ThePrgName = "PolyGraph";

/* World */

void World::syncZipfOmega() {
	// harmonic sum approximation: log(i) + gamma + 1/(2i)
	theZipfOmega = log(theCapacity) + TheEMConst + 1./(2*theCapacity);
}

bool World::loadUrls(const char *fname) {
	char url[4096];

	Assert(fname);

	ifstream is(fname);
	Size sz = 0;
	int lineCount = 0;
	int urlCount = 0;

	SysLog << "loading urls from '" << fname << "'" << endl;

	while (is && is.getline(url, sizeof(url))) {
		lineCount++;
		const int len = strlen(url);
		if (len && *url == '/') {
			theUrls.append(StrDup(url));
			sz += sizeof(char*) + len + 1;
			urlCount++;
		} else
		if (len && *url != '#')
			SysLog << fname << ':' << lineCount << " unknown url format near " << url << endl;
	}

	is.close();

	SysLog
		<< "loaded " << urlCount 
		<< " urls from '" << fname << "'"
		<< " (" << lineCount << " lines)"
		<< " consumed ~ " << sz/(1024*1024.) << " MB" << endl;

	return urlCount > 0;
}

/* Opt */

// set defaults
Opt::Opt(): theValPref(0), theValSuff(0) {

	theOriginAddr = Address("localhost", 8080);
	theProxyAddr = Address("127.0.0.1", 8080);

	theWorldId = 1;
	theWorldCapacity = (1 << 30);
	theSeqStart = -1;  // disabled
	theRndSeed = 1;    // default
	theUniqStart = -1; // disabled

	theConnEstTimout = Time(); // disabled;

	theObjSizeMean = Size::KB(13);

	theUrlsFileName = 0;

	useEphPorts = false;

	theClientCount = 0;      // must be set
	theReqPerClient = -1;    // unlimited default
	theReqTotal = -1;
	theLaunchWindow = Time(0,0); // immediate start
	theClientThinkTime = Time(); // disabled

	theServerCount = 0;
	theSrvWriteDelay = Time(); // disabled
	theIdleTimeout = Time::Sec(5*60); // 5 min default
	isRepCachable = true;
}

Opt::~Opt() {

	delete[] theUrlsFileName;

	delete[] theValPref;
	delete[] theValSuff;
}

bool Opt::parse(int argc, char *argv[]) {
	if (argc < 1+1) /* there is one required option */
		return false;

	int i;
	for (i = 1; i < argc; i += 2) {
		const char *name = argv[i];
		const char *val = argv[i+1];
		if (optEqual(name, "--host")) {
			theOriginAddr = splitVal(val) ?
				Address(theValPref, xatoi(theValSuff)) :
				Address(val, 80);
		} else
		if (optEqual(name, "--proxy") && splitVal(val))
			theProxyAddr = Address(theValPref, xatoi(theValSuff));
		else
		if (optEqual(name, "--world") && val && !isdigit(*val)) {
			theUrlsFileName = StrDup(val);
		} else
		if (optEqual(name, "--world") && splitVal(val) && isdigit(*val)) {
			theWorldId = xatoi(theValPref);
			theWorldCapacity = (strcasecmp(theValSuff, "max")) ?
				xatoi(theValSuff) : (1 << 30);
		} else
		if (optEqual(name, "--seq")) {
			theSeqStart = xatoi(val, -1);
			theRndSeed = -1;
		} else
		if (optEqual(name, "--rnd")) {
			theRndSeed = xatoi(val, 1);
			theSeqStart = -1;
		} else
		if (optEqual(name, "--unique"))
			theUniqStart = xatoi(val, -1);
		else
		if (optEqual(name, "--obj_size"))
			theObjSizeMean = Size::KB(xatoi(val));
		else
		if (optEqual(name, "--conn_est_tout"))
			theConnEstTimout = Time::Secd(xatoi(val, 0)/1000.);
		else
		if (optEqual(name, "--eph_ports"))
			useEphPorts = true;
		else
		if (optEqual(name, "--launch_win"))
			theLaunchWindow = Time::Sec(xatoi(val, 0));
		else
		if (optEqual(name, "--srv_wr_delay"))
			theSrvWriteDelay = strcasecmp(val, "none") ?
				Time::Sec(xatoi(val, -1)) : Time();
		else
		if (optEqual(name, "--rep_cachable"))
			isRepCachable = xatoi(val, 1) > 0;
		else
		if (optEqual(name, "--idle_tout"))
			theIdleTimeout = strcasecmp(val, "none") ?
				Time::Sec(xatoi(val, -1)) : Time();
		else
		if (isdigit(*name))
			break;
		else {
			cerr << "unknown/broken option: '" << name << "'" << endl << endl;
			return false;
		}
	}

	if (!argv[i] || !isdigit(*argv[i]))
		return false;

	// process client specs
	theClientCount = xatoi(argv[i]);
	if (const char *h = strchr(argv[i], 'x')) { /* req/client */
		theReqPerClient = xatoi(h+1);
	}
	if (const char *h = strchr(argv[i], 't')) { /* total req */
		theReqTotal = xatoi(h+1);
	}
	if (const char *h = strchr(argv[i], 'r')) { /* poisson */
        const double req_per_second = xatoi(h+1);
		if (req_per_second > 0)
			theClientThinkTime = Time::Secd(theClientCount/req_per_second);
		else
			return false;
	}
	
	// server specs (listen ports)
	i++;
	theServerCount = argc - i;
	for (int p = 0; p < theServerCount; ++p)
		theServerPorts.append(xatoi(argv[i+p]));

	if (theClientCount > 0)
		return theReqPerClient > 0 || theReqTotal > 0;
	else
		return theServerCount > 0;
}


void Opt::report(ostream &os) const {
	os << "configuration:" << endl
		<< "\tOriginAddr:\t " << theOriginAddr << endl
		<< "\tProxyAddr:\t " << theProxyAddr << endl
		<< "\tWorldId:\t " << theWorldId << endl
		<< "\tWorldCapacity:\t " << theWorldCapacity << " objects" << endl
		<< "\tSeqStart:\t " << theSeqStart << endl
		<< "\tRndSeed:\t " << theRndSeed << endl
		<< "\tUniqStart:\t " << theUniqStart << endl
		<< "\tObjSizeMean:\t " << theObjSizeMean << " bytes" << endl
		<< "\tConnEstTimout:\t " << theConnEstTimout << endl
		<< "\tUseEphPorts:\t " << (useEphPorts ? "yes" : "no") << endl
		<< "\tClientCount:\t " << theClientCount << endl
		<< "\tReqPerClient:\t " << theReqPerClient << endl
		<< "\tReqTotal:\t " << theReqTotal << endl
		<< "\tLaunchWindow:\t " << theLaunchWindow << endl
		<< "\tClntThinkTime:\t " << theClientThinkTime << endl
		<< "\tUrlsFile:\t " << (theUrlsFileName ? theUrlsFileName : "<none>") << endl
		<< "\tServerCount:\t " << theServerCount << endl
		<< "\tSrvWriteDelay:\t " << theSrvWriteDelay << endl
		<< "\tIdleTimeout:\t " << theIdleTimeout << endl
		<< "\tRepCachable:\t " << (isRepCachable ? "yes" : "no") << endl;

	os << "\tServerPorts:";
	for (int p = 0; p < theServerPorts.count(); ++p)
		os << "\t " << setw(4) << theServerPorts[p];
	os << endl;

	os << "workload:   " << (theClientThinkTime < 0 ? "stream" : "poisson") << endl;
	os << "obj_id_gen: " << (theRndSeed > 0 ? "random" : "sequential") << endl;
	os << "unique_url_gen: " << (theUniqStart < 0 ? "disabled" : "enabled") << endl;
}

bool Opt::optEqual(const char *name, const char *canonic) {
	if (!name)
		return false;
	
	// strip leading '-'
	while (*name == '-') name++;
	while (*canonic == '-') canonic++;

	if (!*name)
		return false;

	// matching prefix == equal! (sort of a kludge, be careful)
	return strncasecmp(name, canonic, strlen(name)) == 0;
}

bool Opt::splitVal(const char *value) {
	if (!value)
		return false;
	delete[] theValPref;
	delete[] theValSuff;
	theValPref = StrDup(value);
	if (char *h = strchr(theValPref, ':')) {
		*h = '\0';
		theValSuff = StrDup(h+1);
		return true;
	} else {
		cerr << "column (':') expected in '" << value << "'." << endl;
		theValSuff = StrDup("");
		return false;
	}	
}
